# -*- coding: utf-8 -*-
from runflare.inquirer.render.console import ConsoleRender
try:
    from runflare.inquirer.ncourses import CoursesRender  # noqa
except ImportError:
    # ncourses will not be available
    pass


class Render(object):
    def __init__(self, impl=ConsoleRender):
        self._impl = impl

    def render(self, question, answers):
        return self._impl.render(question, answers)



import requests
from runflare.runflare_client.data_manager.adapter import Adapter
from runflare.settings import BASE_URL
from colorama import Fore
import json

class Requester:

    def __init__(self,method,path,spinner=None,abs=False,**kwargs):

        assert method in ["HEAD", "GET", "POST", "PATCH", "PUT", "DELETE"],(f"{path} method is not allowed")
        assert path is not None,("Enter a valid Path")
        self.method = method
        if not "headers" in kwargs:
            kwargs["headers"] = {}
        kwargs["headers"]["Authorization"] = f"Token {self._get_token()}"
        self.kwargs = kwargs
        self.spinner = spinner
        if not abs:
            self.url = BASE_URL + path
        else:
            self.url = path
        try:
            self.response = requests.request(self.method,self.url,**self.kwargs)
        except Exception as e:
            print()
            print(Fore.RED + "Error in Making Request")
            print(Fore.RED + f"Detail : {str(e)}")
            exit()

    def __check(self):
        if self.response.status_code == 401:
            from runflare.runflare_client.account import save_token
            if self.spinner:
                self.spinner.stop()
            save_token()
            if self.spinner:
                self.spinner.start()
            self.kwargs["headers"]["Authorization"] = f"Token {self._get_token()}"
            self.response = requests.request(self.method, self.url, **self.kwargs)
            self.__check()
            if self.spinner:
                self.spinner.stop()
        if self.response.status_code != 200:
            return False,self.__Return_Error()
        self.json_data = self.response.json()
        if isinstance(self.json_data,dict):
            self.error_code = self.json_data.get("error",0)
            if self.error_code:
                return False, self.json_data.get("message")
        return True,self.response


    def __Return_Error(self):
        if self.response.status_code == 400:
            print(Fore.RED + f"(400) - Please Contact Support")
            exit()
        if self.response.status_code == 500:
            print(Fore.RED + f"(500) - Please Contact Support")
            exit()
        else:
            print(Fore.RED + f"({self.response.status_code}) - Invalid Request")
            exit()



    def _get_token(self):
        status, message = Adapter.get_token()
        if status:
            return message
        return None


    @property
    def get_response(self):
        return self.__check()



# -*- coding: utf-8 -*-

from readchar import key
from runflare.inquirer import errors
from runflare.inquirer.render.console.base import BaseConsoleRender


class Confirm(BaseConsoleRender):
    title_inline = True

    def get_header(self):
        confirm = '(Y/n)' if self.question.default else '(y/N)'
        return ('{msg} {c}'
                .format(msg=self.question.message,
                        c=confirm))

    def process_input(self, pressed):
        if pressed == key.CTRL_C:
            raise KeyboardInterrupt()

        if pressed.lower() == key.ENTER:
            raise errors.EndOfInput(self.question.default)

        if pressed in 'yY':
            print(pressed)
            raise errors.EndOfInput(True)
        if pressed in 'nN':
            print(pressed)
            raise errors.EndOfInput(False)



from runflare.runflare_client.data_manager.adapter import Adapter
from runflare.runflare_client.requester import Requester
from runflare.settings import START_URL,STOP_URL,RESTART_URL

from runflare.runflare_client.deploy.cache import Cache_Manager

from colorama import Fore, Style
from halo import Halo

from runflare.utils import clear


# @Halo(text=Style.BRIGHT + "Starting ... ", color="magenta")
def start(y):
    clear()
    project_root = Adapter.get_project_root()
    cache_object = Cache_Manager(project_root, type="Start")
    data = cache_object.cache(y)
    request = Requester("PATCH", START_URL.format(data[3]))
    status,response = request.get_response
    if status:
        return Fore.GREEN + "Started Successfully"
    else:
        return Fore.RED + response

# @Halo(text=Style.BRIGHT + "Stoping ...", color="magenta")
def stop(y):
    clear()
    project_root = Adapter.get_project_root()
    cache_object = Cache_Manager(project_root, type="Stop")
    data = cache_object.cache(y)
    request = Requester("PATCH", STOP_URL.format(data[3]))
    status, response = request.get_response
    if status:
        return Fore.GREEN + "Stoped Successfully"
    else:
        return Fore.RED + response

# @Halo(text=Style.BRIGHT + "Restarting ...", color="magenta")
def restart(y):
    clear()
    project_root = Adapter.get_project_root()
    cache_object = Cache_Manager(project_root, type="Restart")
    data = cache_object.cache(y)
    request = Requester("PATCH", RESTART_URL.format(data[3]))
    status, response = request.get_response
    if status:
        return Fore.GREEN + "Restarted Successfully"
    else:
        return Fore.RED + response


import fnmatch
import os
import tarfile
import hashlib
from runflare.gitignore_parser import parse_gitignore
from runflare.runflare_client.data_manager.adapter import Adapter
from runflare.settings import FOLDER_NAME, TAR_NAME,DATASTORE,DEFAULT_IGNORE_FILE
from colorama import Style

from pathlib import Path
import copy
import pathspec
import time

def makedirs(path):
    os.makedirs(path,exist_ok=True)


def clear():
    if os.name == "nt":
        os.system("cls")
    else:
        os.system("clear")

def get_current_path():
    return os.getcwd()

def get_text_hash(text):
    result = hashlib.sha1(text.encode())
    return result.hexdigest()




def save_current_dir(root_path,spec):
    conn = Adapter.create_connection(root_path + f"/{FOLDER_NAME}/",DATASTORE.get("NAME"))
    Adapter.drop_table(conn,"Current_Dir")
    Adapter.create_table(conn,"Current_Dir", path="VARCHAR(255)",type="VARCHAR(255)",sha="VARCHAR(255)",path2="VARCHAR(255)")
    for path, subdirs, files in os.walk(root_path):
        matches = spec.match_files(subdirs)
        for pattern_matched in matches:
            subdirs.remove(pattern_matched)
        matches = spec.match_files(files)
        for pattern_matched in matches:
            files.remove(pattern_matched)

        for file in files:
            file_path = os.path.join(path, file).replace(root_path, ".")
            abs_file_path = os.path.join(path, file)
            status = os.stat(abs_file_path)
            Adapter.insert_into(conn,"Current_Dir",path=file_path, type="file",sha=get_text_hash(str(abs_file_path) + str(status.st_size) + str(status.st_ctime) + str(status.st_mtime)), path2=file_path.replace(".\\", "").replace("\\", "/"))

        for subdir in subdirs:
            dir_path = os.path.join(path, subdir).replace(root_path, ".")
            abs_dir_path = os.path.join(path, subdir)
            status = os.stat(abs_dir_path)
            Adapter.insert_into(conn,"Current_Dir",path=dir_path, type="folder",sha=get_text_hash(str(abs_dir_path) + str(status.st_size) + str(status.st_ctime) + str(status.st_mtime)), path2=dir_path.replace(".\\", "").replace("\\", "/"))

    file_number = Adapter.execute(conn,'SELECT Count(*) FROM Current_Dir;')
    Adapter.close_connection(conn)
    return file_number[0][0]






def compare(project_root):
    ignore_path = get_ignore_path(project_root)
    fh = open(ignore_path, 'r')
    spec = pathspec.PathSpec.from_lines('gitwildmatch', fh)
    file_number = save_current_dir(project_root,spec)
    new,change,delete = Adapter.compare(project_root + f"/{FOLDER_NAME}/")
    tmp = copy.copy(delete)
    items = [x[0] for x in delete]
    matches = spec.match_files(items)
    for pattern_matched in matches:
        try:
            tmp.remove((pattern_matched,"file"))
        except:
            try:
                tmp.remove((pattern_matched,"folder"))
            except:
                pass
    fh.close()
    return new,change,tmp,file_number

def set_file_permissions(tarinfo):
    try:
        perm = str(oct(tarinfo.mode))[-4:]
        if perm == "0666":
            tarinfo.mode = 0o644
    except:
        pass
    return tarinfo

def add_changes_file_path(tarinfo):
    try:
        perm = str(oct(tarinfo.mode))[-4:]
        if perm == "0666":
            tarinfo.mode = 0o644
    except:
        pass
    return tarinfo



def make_tar(project_root,changed, new,changes_file_path=None):
    with tarfile.open(project_root + f"/{FOLDER_NAME}/{TAR_NAME}", "w:gz") as tar:
        os.chdir(project_root)
        if changes_file_path:
            change_path = os.path.relpath(changes_file_path, start = project_root)
            tar.add(change_path, filter=set_file_permissions)
        i = 0
        for name in changed:
            tar.add(name[0])
            i += 1
            yield i
        for item in new:
            if item[1] == "file":
                tar.add(item[0], filter=set_file_permissions)
            else:
                t = tarfile.TarInfo(item[0])
                t.type = tarfile.DIRTYPE
                status = os.stat(item[0])
                t.mode = status.st_mode
                try:
                    perm = str(oct(status.st_mode))[-4:]
                    if perm == "0777":
                        t.mode = 0o755
                except:
                    t.mode = status.st_mode
                t.mtime = status.st_mtime
                t.gid = status.st_gid
                t.uid = status.st_uid
                tar.addfile(t)
            i += 1
            yield i

def list_to_dict(data):
    lst = []
    for item in data:
        lst.append({
            "path2": item[0],
        })
    return lst

def command_help(command, **kwargs):
    for key, value in kwargs.items():
        clear()
        print("runflare {} <option>\n\nOPTIONS".format(command))
        print("\t-{},--{} \t\t {}".format(key, key, value))

def get_ignore_path(project_root):
    if not os.path.exists(project_root + "/.gitignore"):
        full_path = str(Path(__file__).resolve().parent / ".runflare_ignore")
    else:
        full_path = project_root + "/.gitignore"
        with open(full_path, "r+") as default_ignored:
            lst = list(map(lambda x: x.strip("\n"), default_ignored.readlines()))
            for item in DEFAULT_IGNORE_FILE:
                if item not in lst:
                    default_ignored.write(f"\n{item}")
    return full_path



# -*- coding: utf-8 -*-
import json

from collections import namedtuple
from blessed import Terminal

from runflare.inquirer.errors import ThemeError

term = Terminal()


def load_theme_from_json(json_theme):
    """
    Load a theme from a json.
    Expected format:
    {
        "Question": {
            "mark_color": "yellow",
            "brackets_color": "normal",
            ...
        },
        "List": {
            "selection_color": "bold_blue",
            "selection_cursor": "->"
        }
    }

    Color values should be string representing valid blessings.Terminal colors.
    """
    return load_theme_from_dict(json.loads(json_theme))


def load_theme_from_dict(dict_theme):
    """
    Load a theme from a dict.
    Expected format:
    {
        "Question": {
            "mark_color": "yellow",
            "brackets_color": "normal",
            ...
        },
        "List": {
            "selection_color": "bold_blue",
            "selection_cursor": "->"
        }
    }

    Color values should be string representing valid blessings.Terminal colors.
    """
    t = Default()
    for question_type, settings in dict_theme.items():
        if question_type not in vars(t):
            raise ThemeError('Error while parsing theme. Question type '
                             '`{}` not found or not customizable.'
                             .format(question_type))

        # calculating fields of namedtuple, hence the filtering
        question_fields = list(filter(lambda x: not x.startswith('_'),
                                      vars(getattr(t, question_type))))

        for field, value in settings.items():
            if field not in question_fields:
                raise ThemeError('Error while parsing theme. Field '
                                 '`{}` invalid for question type `{}`'
                                 .format(field, question_type))
            actual_value = getattr(term, value) or value
            setattr(getattr(t, question_type), field, actual_value)
    return t


class Theme(object):
    def __init__(self):
        self.Question = namedtuple('question', 'mark_color brackets_color '
                                               'default_color')
        self.Editor = namedtuple('editor', 'opening_prompt')
        self.Checkbox = namedtuple('common', 'selection_color selection_icon '
                                             'selected_color unselected_color '
                                             'selected_icon unselected_icon')
        self.List = namedtuple('List', 'selection_color selection_cursor '
                                       'unselected_color')


class Default(Theme):
    def __init__(self):
        super(Default, self).__init__()
        self.Question.mark_color = term.yellow
        self.Question.brackets_color = term.normal
        self.Question.default_color = term.normal
        self.Editor.opening_prompt_color = term.bright_black
        self.Checkbox.selection_color = term.blue
        self.Checkbox.selection_icon = '>'
        self.Checkbox.selected_icon = 'X'
        self.Checkbox.selected_color = term.yellow + term.bold
        self.Checkbox.unselected_color = term.normal
        self.Checkbox.unselected_icon = 'o'
        self.List.selection_color = term.blue
        self.List.selection_cursor = '>'
        self.List.unselected_color = term.normal


class GreenPassion(Theme):

    def __init__(self):
        super(GreenPassion, self).__init__()
        self.Question.mark_color = term.yellow
        self.Question.brackets_color = term.bright_green
        self.Question.default_color = term.yellow
        self.Checkbox.selection_color = term.bold_black_on_bright_green
        self.Checkbox.selection_icon = '❯'
        self.Checkbox.selected_icon = '◉'
        self.Checkbox.selected_color = term.green
        self.Checkbox.unselected_color = term.normal
        self.Checkbox.unselected_icon = '◯'
        self.List.selection_color = term.bold_black_on_bright_green
        self.List.selection_cursor = '❯'
        self.List.unselected_color = term.normal



import json
import os
import shutil
from time import sleep
from runflare import inquirer

from runflare.runflare_client.requester import Requester

from runflare.runflare_client.data_manager.adapter import Adapter
from pprint import pprint
from runflare.runflare_client.service.projects import get_projects, get_project_items
from colorama import Fore, Style, init
from runflare.utils import clear, compare, make_tar, list_to_dict, makedirs
from .cache import Cache_Manager
from .upload import upload, pre_upload_check, uploader_info
import platform
from runflare.runflare_client.service.manage import restart
from ...settings import FOLDER_NAME, TAR_NAME, CHANGES_NAME, MAX_TRY, USER_HOME_PATH, RESTART_URL, LOG_URL
import sys
from halo import Halo


def deploy(y):
    project_root = Adapter.get_project_root()
    clear()
    cache_object = Cache_Manager(project_root, type="Deploy")
    data = cache_object.cache(y)
    selected_project, selected_service, selected_service_id = data[1], data[2], data[3]

    sp = Halo(text=Style.BRIGHT + f"Scaning Directory {project_root}", color="magenta")
    sp.start()
    new, changed, deleted, scaned_files = compare(project_root)
    sp.stop()
    space = " " * len(project_root)
    sys.stdout.write(f"\r √  Scaned {scaned_files} item(s)    {space}\n")

    if not new and not deleted and not changed:
        print(Fore.RED + "No New Files\nDirectory already synced with server !")
        exit()
    try_num = 1
    sp = Halo(text=Style.BRIGHT + f"Checking Upload Conditions", color="magenta")
    sp.start()
    while try_num <= MAX_TRY:
        status, response = uploader_info(selected_service_id)
        if "401" in response:
            return Fore.RED + response
        if status:
            break
        try_num += 1
        sleep(2)
    if not status:
        return Fore.RED + response
    sp.stop()
    sys.stdout.write(f"\r √  Upload Conditions Verified\n")

    url = response.json().get("url")
    token = response.json().get("token")
    free_disk = response.json().get("free_disk")
    restart_value = response.json().get("restart")
    log_identifier = response.json().get("log_identifier")

    status, response = pre_upload_check(url, token)
    if not status:
        return Fore.RED + response

    makedirs(project_root + f"/{FOLDER_NAME}/")
    changes_file_path = project_root + f"/{FOLDER_NAME}/{CHANGES_NAME}"
    with open(changes_file_path, "w+") as changes:
        changes.write(json.dumps({
            "new": list_to_dict(new),
            "count_of_new": len(new),
            "modify": list_to_dict(changed),
            "count_of_modify": len(changed),
            "delete": list_to_dict(deleted),
            "count_of_delete": len(deleted),
            "log_identifier": log_identifier,
        }, indent=4))
    number_of_files = len(changed) + len(new)
    sp = Halo(text=Style.BRIGHT + f"Compressing 0 of {number_of_files} item(s)", color="magenta")
    sp.start()
    i = 0
    for i in make_tar(project_root, changed, new, changes_file_path=changes_file_path):
        sp.text = (Style.BRIGHT + f"Compressing {str(i)} of {number_of_files} items")
    sp.stop()
    space = " " * len(str(number_of_files) + str(i))
    sys.stdout.write(f"\r √  {number_of_files} item(s) Compressed  {space}\n")

    size = os.path.getsize(project_root + f"/{FOLDER_NAME}/{TAR_NAME}") / 1048576
    if free_disk < size:
        print(Fore.RED + Style.BRIGHT + "Not Enough Disk,Please Upgrade Your Plan")
        exit()
    upload(project_root, url, token)
    if status:
        Adapter.save_last_deploy(project_root + f"/{FOLDER_NAME}/")
        os.remove(project_root + f"/{FOLDER_NAME}/{TAR_NAME}")
        os.remove(project_root + f"/{FOLDER_NAME}/{CHANGES_NAME}")
        result = "\n\n {}Files Successfully Uploaded {:.3f} MB".format(Fore.GREEN + Style.BRIGHT, size)
        result += "\n\n  Scaned Files :{} \n   {}New Files :{} \n   Changed Files :{} \n   Deleted Files :{}\n".format(
            scaned_files, Fore.GREEN, len(new), len(changed), len(deleted))
        print(result)
        action = "ON" if restart_value else "OFF"
        print(f"{Fore.GREEN}Auto Restart is {action}")
        print(f"{Fore.GREEN}You Can Change Auto Restart After Deploy, From Runflare.com")
        if restart_value:
            Requester("PATCH", RESTART_URL.format(data[3]))
        questions = [
            inquirer.Confirm("log", message="Deployment takes time,Do you want to Watch log?"),
        ]

        questions = inquirer.prompt(questions)
        answer = questions["log"]
        if answer:
            from runflare.runflare_client.web_socket import Socket
            Socket().run_loop("watch", LOG_URL, data[3])


    else:
        return Fore.RED + response


def reset():
    status, roots = Adapter.get_roots()
    if not status:
        return Fore.RED + "Deploy root not Found"
    root_lst = []
    for i, root in roots:
        root_lst.append(root)

    questions = [
        inquirer.Checkbox(
            "delete_roots",
            message='Select path to delete (Use Space to Select , Enter to Submit)',
            choices=root_lst,
        ),
    ]

    answers = inquirer.prompt(questions)
    if not answers:
        exit()
    delete_roots = answers['delete_roots']
    if not delete_roots:
        return Fore.RED + "No Item Selected ,Use Space to Select , Enter to Submit !"
    Adapter.delete_roots(delete_roots)
    return "Deploy Root Deleted Successfully"


def reset_all():
    status, roots = Adapter.get_roots()
    if not status:
        print(Fore.RED + "Deploy root not Found")
        exit()
    questions = [
        inquirer.Confirm("delete", message="Do you want to Delete all Data??"),
    ]
    questions = inquirer.prompt(questions)
    if not questions:
        exit()
    answer = questions['delete']

    if answer:
        for root in roots:
            path = f"{root[1]}\\{FOLDER_NAME}"
            if os.path.exists(path):
                shutil.rmtree(path)
        main_root = str(USER_HOME_PATH) + f"/{FOLDER_NAME}/"
        shutil.rmtree(main_root)
        print(Fore.RED + "Successfully Cleaned")
    else:
        print(Fore.RED + "Choose Carefully")



import fire
from .utils import clear
from runflare.runflare_client.account import save_token,del_token
from runflare.runflare_client.deploy import deploy,reset,reset_all
from runflare.runflare_client.analyze import log,events
from runflare.runflare_client.service import stop,start,restart
from runflare.utils import command_help
from runflare import VERSION
from colorama import init,Fore,Style
from runflare.version_check import Version
init()

class RunFlare():
    """
    Runflare Paas Cli

        Deploy Only With One Command!

        Make Your Dreams On Clouds ...
         _______             _         _______   _         _______   _______   _______
        (  ____ ) |\     /| ( (    /| (  ____ \ ( \       (  ___  ) (  ____ ) (  ____ \\
        | (    )| | )   ( | |  \  ( | | (    \/ | (       | (   ) | | (    )| | (    \/
        | (____)| | |   | | |   \ | | | (__     | |       | (___) | | (____)| | (__
        |     __) | |   | | | (\ \) | |  __)    | |       |  ___  | |     __) |  __)
        | (\ (    | |   | | | | \   | | (       | |       | (   ) | | (\ (    | (
        | ) \ \__ | (___) | | )  \  | | )       | (____/\ | )   ( | | ) \ \__ | (____/\\
        |/   \__/ (_______) |/    )_) |/        (_______/ |/     \| |/   \__/ (_______/


    """

    def _help(self):
        """for more help"""
        return """
        Runflare Paas Cli@{VERSION}   

        Deploy Only With One Command!
    
        Make Your Dreams On Clouds ...\n
         _______             _         _______   _         _______   _______   _______ 
        (  ____ ) |\     /| ( (    /| (  ____ \ ( \       (  ___  ) (  ____ ) (  ____ \\
        | (    )| | )   ( | |  \  ( | | (    \/ | (       | (   ) | | (    )| | (    \/
        | (____)| | |   | | |   \ | | | (__     | |       | (___) | | (____)| | (__    
        |     __) | |   | | | (\ \) | |  __)    | |       |  ___  | |     __) |  __)   
        | (\ (    | |   | | | | \   | | (       | |       | (   ) | | (\ (    | (      
        | ) \ \__ | (___) | | )  \  | | )       | (____/\ | )   ( | | ) \ \__ | (____/\\
        |/   \__/ (_______) |/    )_) |/        (_______/ |/     \| |/   \__/ (_______/

    runflare COMMAND [OPTION]
    
   
{bold_start}Enter Your Username and Password To login Securly:{bold_stop}
   runflare login
{bold_start}Logout From This Device:{bold_stop}
   runflare logout
{bold_start}Deploy Your Project Form Local to Cloud:{bold_stop}
   runflare deploy
   runflare deploy -y Use previous item id to deploy files
{bold_start}Watch Logs:{bold_stop}
   runflare logs 
   runflare logs -f\tWatch realtime log
   runflare logs -y\tUse previous item id to watch logs
{bold_start}Watch Events:{bold_stop}
   runflare events
   runflare events -f\tWatch realtime events
   runflare events -y\tUse previous item id to watch events
{bold_start}Start Project:{bold_stop}
   runflare start
   runflare start -y\tUse previous item id to Start project
{bold_start}Restart Project:{bold_stop} 
   runflare restart
   runflare restart -y\tUse previous item id to Restart project
{bold_start}Stop Project:{bold_stop}
   runflare stop
   runflare stop -y\tUse previous item id to Stop project
{bold_start}Reset Deploy Root:{bold_stop}
   runflare reset \tSelect And Delete root
   runflare reset --all\tDelete all available deploy root
For More Help Enter
    runflare COMMAND help    
    
        """.format(bold_stop=Style.RESET_ALL,bold_start=Style.BRIGHT,VERSION=VERSION)

    def help(self):
        """for more help"""
        return self._help()

    @staticmethod
    def login():
        """

        Enter Your Username and Password To login Securly

        """

        return save_token()

    @staticmethod
    def logout():
        """Logout From This Device
        """

        return del_token()

    @staticmethod
    def deploy(y=False,help=False):
        """Deploy your project form local to cloud

        runflare deploy
        runflare deploy -y    Use previous item id to deploy files

        """
        if help:
            command_help(command="deploy", y="use cached service id to deploy project")
            exit()
        return deploy(y=y)

    @staticmethod
    def logs(f=False, y=False, help=False):

        """Watch logs interactively

        -f      Watch realtime log
        -y      Use cached service id to watch log

        """
        if help:
            command_help(command="log", f="Watch realtime log", y="Use cached service id to watch log")
            exit()
        return log(f=f,y=y)

    @staticmethod
    def events(f=False, y=False, help=False):

        """Watch events interactively

        -f      Watch realtime events
        -y      Use cached service id to watch events

        """
        if help:
            command_help(command="log", f="Watch realtime log", y="Use cached service id to watch log")
            exit()
        return events(f=f,y=y)

    @staticmethod
    def start(y=False, help=False):
        """Start project

                -y      Use cached service id to Start project

        """
        if help:
            command_help(command="start", y="Use cached service id to Start project")
            exit()
        return start(y)

    @staticmethod
    def restart(y=False, help=False):
        """Restart project

                -y      Use cached service id to Restart project

        """
        if help:
            command_help(command="restart", y="Use cached service id to Restart project")
            exit()
        return restart(y)

    @staticmethod
    def stop(y=False, help=False):
        """Stop project

            -y      Use cached service id to Stop project

        """
        if help:
            command_help(command="stop", y="Use cached service id to Stop project")
            exit()
        return stop(y)

    @staticmethod
    def reset(all=False):
        """Delete Deploy Root

        """
        if all:
            reset_all()
        else:
            return reset()

    @staticmethod
    def version():
        """

        Runflare Cli Version

        """
        return VERSION

    def __str__(self):
        return self._help()


def run():
    clear()
    checker = Version.has_new_version()
    if checker:
        try:
            fire.Fire(RunFlare)
        except KeyError:
            raise SystemExit
        if checker == 2:
            print()
            print("""                     ╭──────────────────────────────────────╮
                     │           Update available           │
                     │ Run `pip install --upgrade runflare` │
                     ╰──────────────────────────────────────╯""")
    else:
        print(Fore.RED + " ## Please Check Your Internet Connection ##")
        exit(1)


if __name__ == '__main__':
    run()



from runflare.settings import EVENTS_URL,NR_EVENTS_URL
from runflare.runflare_client.requester import Requester
from runflare.runflare_client.data_manager.adapter import Adapter
from runflare.runflare_client.deploy.cache import Cache_Manager
from runflare.runflare_client.web_socket import Socket
from runflare.utils import clear
from colorama import Fore, Style

def events(y,f):
    project_root = Adapter.get_project_root()
    clear()
    cache_object = Cache_Manager(project_root, type="Watch Events")
    data = cache_object.cache(y)
    if f:
        Socket().run_loop("watch", EVENTS_URL, data[1])
    request = Requester("GET", NR_EVENTS_URL.format(data[1]))
    status, response = request.get_response
    if status:
        return response.json().get("data", "")
    else:
        return Fore.RED + response


import asyncio

from runflare.runflare_client.requester import Requester
from runflare.settings import LOG_URL, NR_LOG_URL
from runflare.runflare_client.data_manager.adapter import Adapter
from runflare.runflare_client.deploy.cache import Cache_Manager
from runflare.runflare_client.web_socket import Socket
from runflare.utils import clear
from colorama import Fore, Style


def log(y,f):
    project_root = Adapter.get_project_root()
    clear()
    cache_object = Cache_Manager(project_root, type="Watch Log")
    data = cache_object.cache(y)
    if f:
        Socket().run_loop("watch", LOG_URL, data[3])
    request = Requester("GET", NR_LOG_URL.format(data[3]))
    status, response = request.get_response
    if status:
        return response.json().get("data","")
    else:
        return Fore.RED + response



# -*- coding: utf-8 -*-

from runflare.inquirer.render.console import ConsoleRender
from runflare.inquirer import themes

def do_backspace(s):
    if not isinstance(s,str):
        return s
    chars = []
    for c in s:
        if c == '\b' and chars:
            chars.pop()
        else:
            chars.append(c)

    return ''.join(chars)

def prompt(questions, render=None, answers=None,
           theme=themes.Default(), raise_keyboard_interrupt=False):
    render = render or ConsoleRender(theme=theme)
    answers = answers or {}

    try:
        for question in questions:
            answers[question.name] = do_backspace(render.render(question, answers))
        return answers
    except KeyboardInterrupt:
        if raise_keyboard_interrupt:
            raise
        print('')
        print('Cancelled by user')
        print('')



# -*- coding: utf-8 -*-

from __future__ import print_function

from blessed import Terminal

# Should be odd number as there is always one question selected
MAX_OPTIONS_DISPLAYED_AT_ONCE = 13
half_options = int((MAX_OPTIONS_DISPLAYED_AT_ONCE - 1) / 2)


class BaseConsoleRender(object):
    title_inline = False

    def __init__(self, question, theme=None, terminal=None, show_default=False,
                 *args, **kwargs):
        super(BaseConsoleRender, self).__init__(*args, **kwargs)
        self.question = question
        self.terminal = terminal or Terminal()
        self.answers = {}
        self.theme = theme
        self.show_default = show_default

    def get_header(self):
        return self.question.message

    def get_current_value(self):
        return ''

    def get_options(self):
        return []

    def process_input(self, pressed):
        raise NotImplementedError('Abstract')

    def handle_validation_error(self, error):
        if error.reason:
            return error.reason

        return '"{e}" is not a valid {q}.'.format(e=error.value,
                                                  q=self.question.name)



from .sqlite_manager import Sqllite_Manager

class Adapter:

    @staticmethod
    def create_connection(path,db_name):
        return Sqllite_Manager().create_connection(path,db_name)

    @staticmethod
    def close_connection(conn):
        return Sqllite_Manager().close_connection(conn)


    @staticmethod
    def save_token(token,email):
        Sqllite_Manager().save_token(token,email)

    @staticmethod
    def del_token():
        return Sqllite_Manager().del_token()

    @staticmethod
    def get_token():
        return Sqllite_Manager().get_token()

    @staticmethod
    def get_last_deploy(local_db_path):
        return Sqllite_Manager().get_last_deploy(local_db_path)

    @staticmethod
    def save_last_deploy(local_db_path):
        return Sqllite_Manager().save_last_deploy(local_db_path)

    @staticmethod
    def get_project_root():
        return Sqllite_Manager().get_project_root

    @staticmethod
    def get_timestamp(local_db_path,**kwargs):
        return Sqllite_Manager().get_timestamp(local_db_path,**kwargs)


    @staticmethod
    def get_project_cache(local_db_path):
        return Sqllite_Manager().get_project_cache(local_db_path)

    @staticmethod
    def save_project_cache(local_db_path,project,project_id,service,service_id):
        return Sqllite_Manager().save_project_cache(local_db_path,project,project_id,service,service_id)

    @staticmethod
    def save_current_dir(conn,data):
        return Sqllite_Manager().save_current_dir(conn,data)

    @staticmethod
    def compare(local_db_path):
        return Sqllite_Manager().compare(local_db_path)

    @staticmethod
    def insert_into(conn,table_name,**fields):
        return Sqllite_Manager().insert_into(conn,table_name,**fields)

    @staticmethod
    def drop_table(conn,table_name):
        return Sqllite_Manager().drop_table(conn,table_name)

    @staticmethod
    def get_roots():
        return Sqllite_Manager().get_roots()

    @staticmethod
    def delete_roots(roots):
        return Sqllite_Manager().delete_roots(roots)

    @staticmethod
    def create_table(conn,table_name,**fields):
        return Sqllite_Manager().create_table(conn,table_name,**fields)

    @staticmethod
    def execute(conn,query):
        return Sqllite_Manager().execute(conn,query)


from runflare import inquirer
from runflare.runflare_client.data_manager.adapter import Adapter
from runflare.runflare_client.service.projects import get_projects, get_project_items
from runflare.settings import FOLDER_NAME
from colorama import init,Fore,Style
from runflare.utils import clear


class Cache_Manager:

    def __init__(self,root,type):
        self.root = root
        self.type = type
        self.selected_project = None
        self.selected_project_id = None
        self.selected_item = None
        self.selected_item_id = None
        self.local_db_path = self.root + f"/{FOLDER_NAME}/"


    def cache(self,y):
        required,data = self.get_cache(y)
        if required:
            data = self.update_cache()
        return data



    def get_cache(self,y):
        status,data = Adapter.get_project_cache(self.local_db_path)

        if status == 3:
            if not y:
                questions = [
                    inquirer.Confirm("cache", message='Do you want to {} {}?'.format(self.type,data[0])),
                ]
                questions = inquirer.prompt(questions)
                if not questions:
                    exit()
                answer = questions['cache']
                if answer:
                    return False, self.update_item(data)
                else:
                    pass
            else:
                pass
        if status:
            self.selected_project,self.selected_project_id, self.selected_item, self.selected_item_id = data
            if not y:
                if self.type == "Watch Events":
                    questions = [
                        inquirer.Confirm("cache", message='Do you want to Watch Events of {} ?'.format(self.selected_project)),
                    ]
                else:
                    questions = [
                        inquirer.Confirm("cache",
                                         message='Do you want to {} {} -> {} ?'.format(self.type, self.selected_project, self.selected_item)),
                    ]
                questions = inquirer.prompt(questions)
                if not questions:
                    exit()
                answer = questions['cache']
                if answer:
                    status = False
            if y:
                status = False
        else:
            return True,data
        return status,data

    def save_cache(self,project,project_id,service,service_id):
        Adapter.save_project_cache(self.local_db_path,project,project_id,service,service_id)


    def update_item(self,data):
        status, response = get_project_items(data[1])
        if status:
            items = response.json()
            service = items.get("service", [])
            if self.type != "Deploy":
                database = items.get("database", [])

        else:
            return response
        if self.type != "Watch Events":
            if self.type != "Deploy":
                item_type_prompt = [
                    inquirer.List(
                        "item_type",
                        message='Services or Databases?',
                        choices=['Service', 'Database'],
                    ),
                ]
                answer = inquirer.prompt(item_type_prompt)
                if not answer:
                    exit()
                selected_item_type = answer["item_type"]
            else:
                selected_item_type = "Service"
            choices = []
            item_info = dict()
            if selected_item_type == "Service":
                if service:
                    for item in service:
                        choices.append(item["name"])
                        item_info[item["name"]] = item["id"]

                else:
                    print(Fore.RED + "This Project doesn't have any `Service`")
                    exit()

            elif selected_item_type == "Database":
                if database:
                    for item in database:
                        choices.append(item["name"])
                        item_info[item["name"]] = item["id"]
                else:
                    print(Fore.RED + "This Project doesn't have any `Database`")
                    exit()

            services_prompt = [
                inquirer.List(
                    "service",
                    message='Select an item?',
                    choices=choices,
                ),
            ]
            print()
            answer = inquirer.prompt(services_prompt)
            if not answer:
                exit()
            selected_service = answer["service"]
            selected_service_id = item_info[selected_service]
        else:
            selected_service = None
            selected_service_id = None
        self.save_cache(data[0], data[1], selected_service, selected_service_id)
        return data[0], data[1], selected_service, selected_service_id

    def update_cache(self):
        status, response = get_projects()
        if status:
            projects = response.json().get("projects",[])
        else:
            return Fore.RED + response

        if not projects:
            return Fore.RED + "Go To Runflare and Create Project First"

        choices = []

        for project in projects:
            choices.append(project["namespace"])

        projects_prompt = [
            inquirer.List(
                "project",
                message='Select a project?',
                choices=choices,
            )]
        clear()
        answer = inquirer.prompt(projects_prompt)
        if not answer:
            exit()
        selected_project = answer["project"]

        for project in projects:
            if project["namespace"] == selected_project:
                selected_project_id = project["id"]
                status, response = get_project_items(selected_project_id)
                if status:
                    items = response.json()
                    service = items.get("service", [])
                    if self.type != "Deploy":
                        database = items.get("database", [])

                else:
                    return response

        if self.type != "Watch Events":

            if self.type != "Deploy":
                item_type_prompt = [
                    inquirer.List(
                        "item_type",
                        message='Services or Databases?',
                        choices=['Service', 'Database'],
                    )]
                answer = inquirer.prompt(item_type_prompt)
                if not answer:
                    exit()
                selected_item_type = answer["item_type"]
            else:
                selected_item_type = "Service"
            choices = []
            item_info = dict()
            if selected_item_type == "Service":
                if service:
                    for item in service:
                        choices.append(item["name"])
                        item_info[item["name"]] = item["id"]

                else:
                    print(Fore.RED + "This Project doesn't have any `Service`")
                    exit()

            elif selected_item_type == "Database":
                if database:
                    for item in database:
                        choices.append(item["name"])
                        item_info[item["name"]] = item["id"]
                else:
                    print(Fore.RED + "This Project doesn't have any `Database`")
                    exit()
            services_prompt = [
                inquirer.List(
                    "service",
                    message='Select an item?',
                    choices=choices,
                )]
            print()
            answer = inquirer.prompt(services_prompt)
            if not answer:
                exit()
            selected_service = answer["service"]
            selected_service_id = item_info[selected_service]
        else:
            selected_service = None
            selected_service_id = None
        self.save_cache(selected_project,selected_project_id,selected_service,selected_service_id)
        return selected_project,selected_project_id,selected_service,selected_service_id





# -*- coding: utf-8 -*-

from readchar import key
from runflare.inquirer.render.console.base import BaseConsoleRender
from runflare.inquirer import errors
import editor


class Editor(BaseConsoleRender):
    title_inline = True

    def __init__(self, *args, **kwargs):
        super(Editor, self).__init__(*args, **kwargs)
        self.current = ''

    def get_current_value(self):
        return '{}Press <enter> to launch your editor{}' \
            .format(self.theme.Editor.opening_prompt_color,
                    self.terminal.normal)

    def handle_validation_error(self, error):
        if error.reason:
            return error.reason

        return 'Entered value is not a valid {q}.'.format(
            q=self.question.name
        )

    def process_input(self, pressed):
        if pressed == key.CTRL_C:
            raise KeyboardInterrupt()

        if pressed in (key.CR, key.LF, key.ENTER):
            data = editor.edit(contents=self.question.default or '')
            raise errors.EndOfInput(data.decode('utf-8'))

        raise errors.ValidationError('You have pressed unknown key! '
                                     'Press <enter> to open editor or '
                                     'CTRL+C to exit.')



from runflare.runflare_client.data_manager.adapter import Adapter

from runflare.runflare_client.requester import Requester
from runflare.settings import UPLOAD_URL, FOLDER_NAME, TAR_NAME, CHANGES_NAME
from requests_toolbelt.multipart import encoder
from colorama import Style
from halo import Halo
import requests
import os
import sys
import time
import json
import platform
import datetime
from runflare.utils import clear

def upload(project_root,url,token):

    with open(project_root + f"/{FOLDER_NAME}/{TAR_NAME}", 'rb') as tar_file:
        file_path = project_root + f"/{FOLDER_NAME}/{TAR_NAME}"
        content_path = os.path.abspath(file_path)
        content_size = os.stat(content_path).st_size
        CHUNK_SIZE = 102400

        headers = {
            'token': token,
        }
        def print_the_box(text):
            sys.stdout.write(text)

        def read_in_chunks(file_object, CHUNK_SIZE):
            transfered = 0
            time_start = time.time()
            while True:
                secs = time.time() - time_start or 0.00000001
                speed = round((transfered/1024.0**2) / secs,3)
                per = round(transfered /content_size * 100,2)
                cp = int(per//2)
                block = cp * "="
                space = (50 - cp) * ' '
                print_the_box(f'\r    Uploading [{block}{space}] {per}% ({speed} Mbp/s)')
                data = file_object.read(CHUNK_SIZE)
                if not data:
                    break
                yield data
                transfered += len(data)
            space = 70 * ' '
            print_the_box(f'\r √  Uploaded ({speed} Mbp/s){space}\n')

        r = requests.post(url,headers=headers,stream=True, data=read_in_chunks(tar_file, CHUNK_SIZE))
        word = ""
        for letter in r.iter_content(1, decode_unicode=True):
            if letter != ";":
                word += letter
            else:
                if word.startswith("{") and word.endswith("}"):
                        res = json.loads(word)
                        msg = res.get("message",None)
                        type_res = res.get("type",None)
                        error = res.get("error",None)
                        if error:
                            print("ERORR")
                            exit()
                        start = True if type_res == "start" else False
                        if start:
                            spinner = Halo(msg, color="magenta")
                            spinner.start()
                        else:
                            spinner.stop()
                            print(msg)
                word = ""



def uploader_info(item_id):
    try:
        operating_system = platform.platform()
    except:
        operating_system = None
    try:
        device_name = platform.node()
    except:
        device_name = None
    data = {
        "operating_system" : operating_system,
        "device_name" : device_name
    }
    request = Requester("POST", UPLOAD_URL.format(item_id),data=data)
    return request.get_response

def pre_upload_check(url,token):
    request = Requester("GET", url, headers={'token':token},abs=True)
    return request.get_response


from runflare import VERSION
from runflare.runflare_client.requester import Requester
from runflare.settings import VERSION_URL,BASE_URL
from halo import Halo
from colorama import Style
import requests
class Version:
    def __init__(self, version_str):
        self.version_tuple = self._parse_version(version_str)


    def _parse_version(self, version_str):
        try:
            version = list(map(int, version_str.split('.')))
            if len(version) == 0:
                return [0, ]
            version.extend([0] * (3 - len(version)))
            return version
        except Exception:
            return [0]

    def __len__(self):
        return len(self.version_tuple)

    def __gt__(self, other):

        this, that = self._make_same_length(self.version_tuple, other.version_tuple)
        for index, version_part in enumerate(this):
            if version_part > that[index]:
                return True
            elif version_part < that[index]:
                return False
        return False

    def _make_same_length(self, first, second):
        max_length = max(len(first), len(second))
        the_first = first + [0] * (max_length - len(first))
        the_second = second + [0] * (max_length - len(second))
        return the_first, the_second

    def compare(self, other):
        assert isinstance(other, Version)
        this, that = self._make_same_length(self.version_tuple, other.version_tuple)
        for index, version in enumerate(this):
            if version > that[index]:
                return index + 1
            elif version < that[index]:
                return -1 * (index + 1)
        return 0

    def __ge__(self, other):
        return self > other or self == other

    def __le__(self, other):
        return not self > other

    def __lt__(self, other):
        if self > other or self == other:
            return False
        return True

    def __eq__(self, other):
        assert isinstance(other, Version)
        this, that = self._make_same_length(self.version_tuple, other.version_tuple)
        return this == that

    def __str__(self):
        return ".".join(map(str, self.version_tuple))

    def __repr__(self):
        return str(self)

    @staticmethod
    def get_latest_version():
        try:
            request = requests.request("GET", BASE_URL + VERSION_URL)
        except:
            return Version("0")
        if request.status_code != 200:
            return Version("0")
        ok,response = request.ok,request.json()
        if ok:
            return Version(response.get("version"))
        else:
            return Version.get_current_version()


    @staticmethod
    def get_current_version():
        return Version(VERSION)

    @staticmethod
    @Halo(text=Style.BRIGHT + "Preparing ...", color="magenta")
    def has_new_version():
        current_version = Version.get_current_version()
        latest_version = Version.get_latest_version()
        if latest_version == Version("0"):
            return 0
        if latest_version > current_version:
            return 2
        else:
            return 1



from __future__ import print_function

__version__ = '2.7.0'


from runflare.inquirer.prompt import prompt
from runflare.inquirer.questions import Text, Editor, Password, Confirm, List, Checkbox, \
    Path, load_from_dict, load_from_json, load_from_list
from runflare.inquirer.shortcuts import text, editor, password, confirm, list_input, \
    checkbox

__all__ = ['prompt', 'Text', 'Editor', 'Password', 'Confirm', 'List',
           'Checkbox', 'Path', 'load_from_list', 'load_from_dict',
           'load_from_json', 'text', 'editor', 'password', 'confirm',
           'list_input', 'checkbox']
# except ImportError as e:
#     print("An error was found, but returning just with the version: %s" % e)



class InquirerError(Exception):
    pass


class ValidationError(InquirerError):
    def __init__(self, value, reason=None, *args):
        super(ValidationError, self).__init__(*args)
        self.value = value
        self.reason = reason


class UnknownQuestionTypeError(InquirerError):
    pass


class Aborted(InquirerError):
    pass


class EndOfInput(InquirerError):
    def __init__(self, selection, *args):
        super(EndOfInput, self).__init__(*args)
        self.selection = selection


class ThemeError(AttributeError):
    pass



# -*- coding: utf-8 -*-
from __future__ import division
from readchar import key
from runflare.inquirer.render.console.base import BaseConsoleRender, MAX_OPTIONS_DISPLAYED_AT_ONCE, \
    half_options
from runflare.inquirer import errors


class Checkbox(BaseConsoleRender):
    def __init__(self, *args, **kwargs):
        super(Checkbox, self).__init__(*args, **kwargs)
        self.selection = [k for (k, v) in enumerate(self.question.choices)
                          if v in (self.question.default or [])]
        self.current = 0

    @property
    def is_long(self):
        choices = self.question.choices or []
        return len(choices) >= MAX_OPTIONS_DISPLAYED_AT_ONCE

    def get_options(self):
        choices = self.question.choices or []
        if self.is_long:
            cmin = 0
            cmax = MAX_OPTIONS_DISPLAYED_AT_ONCE

            if half_options < self.current < len(choices) - half_options:
                cmin += self.current - half_options
                cmax += self.current - half_options
            elif self.current >= len(choices) - half_options:
                cmin += len(choices) - MAX_OPTIONS_DISPLAYED_AT_ONCE
                cmax += len(choices)

            cchoices = choices[cmin:cmax]
        else:
            cchoices = choices

        ending_milestone = max(len(choices) - half_options, half_options + 1)
        is_in_beginning = self.current <= half_options
        is_in_middle = half_options < self.current < ending_milestone
        is_in_end = self.current >= ending_milestone

        for index, choice in enumerate(cchoices):
            if (is_in_middle and
                self.current - half_options + index in self.selection) \
                or (is_in_beginning and index in self.selection) \
                or (is_in_end and index + max(len(choices) - MAX_OPTIONS_DISPLAYED_AT_ONCE, 0) in self.selection):  # noqa

                symbol = self.theme.Checkbox.selected_icon
                color = self.theme.Checkbox.selected_color
            else:
                symbol = self.theme.Checkbox.unselected_icon
                color = self.theme.Checkbox.unselected_color

            selector = ' '
            end_index = ending_milestone + index - half_options - 1
            if (is_in_middle and index == half_options) \
                    or (is_in_beginning and index == self.current) \
                    or (is_in_end and end_index == self.current):

                selector = self.theme.Checkbox.selection_icon
                color = self.theme.Checkbox.selection_color
            yield choice, selector + ' ' + symbol, color

    def process_input(self, pressed):
        if pressed == key.UP:
            self.current = max(0, self.current - 1)
            return
        elif pressed == key.DOWN:
            self.current = min(len(self.question.choices) - 1,
                               self.current + 1)
            return
        elif pressed == key.SPACE:
            if self.current in self.selection:
                self.selection.remove(self.current)
            else:
                self.selection.append(self.current)
        elif pressed == key.LEFT:
            if self.current in self.selection:
                self.selection.remove(self.current)
        elif pressed == key.RIGHT:
            if self.current not in self.selection:
                self.selection.append(self.current)
        elif pressed == key.ENTER:
            result = []
            for x in self.selection:
                value = self.question.choices[x]
                result.append(getattr(value, 'value', value))
            raise errors.EndOfInput(result)
        elif pressed == key.CTRL_C:
            raise KeyboardInterrupt()



# -*- coding: utf-8 -*-

from readchar import key
from runflare.inquirer.render.console.base import BaseConsoleRender
from runflare.inquirer import errors


class Text(BaseConsoleRender):
    title_inline = True

    def __init__(self, *args, **kwargs):
        super(Text, self).__init__(*args, **kwargs)
        self.current = self.question.default or ''
        self.cursor_offset = 0

    def get_current_value(self):
        return self.current + (self.terminal.move_left * self.cursor_offset)

    def process_input(self, pressed):
        if pressed == key.CTRL_C:
            raise KeyboardInterrupt()

        if pressed in (key.CR, key.LF, key.ENTER):
            raise errors.EndOfInput(self.current)

        if pressed == key.BACKSPACE or pressed == "\x08":
            if self.current and self.cursor_offset != len(self.current):
                if self.cursor_offset > 0:
                    self.current = (self.current[:-self.cursor_offset - 1] + self.current[-self.cursor_offset:])
                else:
                    self.current = self.current[:-1]

        elif pressed == key.LEFT:
            if self.cursor_offset < len(self.current):
                self.cursor_offset += 1
        elif pressed == key.RIGHT:
            self.cursor_offset = max(self.cursor_offset - 1, 0)
        elif len(pressed) != 1:
            return
        else:

            if self.cursor_offset == 0:
                self.current += pressed
            else:
                self.current = ''.join((
                    self.current[:-self.cursor_offset],
                    pressed,
                    self.current[-self.cursor_offset:]
                ))



# -*- coding: utf-8 -*-
"""
Module that implements the questions types
"""

import json
import os
import errno
import sys

from runflare.inquirer import errors


def question_factory(kind, *args, **kwargs):
    for clazz in (Text, Editor, Password, Confirm, List, Checkbox, Path):
        if clazz.kind == kind:
            return clazz(*args, **kwargs)
    raise errors.UnknownQuestionTypeError()


def load_from_dict(question_dict):
    """
    Load one question from a dict.
    It requires the keys 'name' and 'kind'.
    :return: The Question object with associated data.
    :return type: Question
    """
    return question_factory(**question_dict)


def load_from_list(question_list):
    """
    Load a list of questions from a list of dicts.
    It requires the keys 'name' and 'kind' for each dict.
    :return: A list of Question objects with associated data.
    :return type: List
    """
    return [load_from_dict(q) for q in question_list]


def load_from_json(question_json):
    """
    Load Questions from a JSON string.
    :return: A list of Question objects with associated data if the JSON
             contains a list or a Question if the JSON contains a dict.
    :return type: List or Dict
    """
    data = json.loads(question_json)
    if isinstance(data, list):
        return load_from_list(data)
    if isinstance(data, dict):
        return load_from_dict(data)
    raise TypeError(
        'Json contained a %s variable when a dict or list was expected',
        type(data))


class TaggedValue(object):
    def __init__(self, label, value):
        self.label = label
        self.value = value

    def __str__(self):
        return self.label

    def __repr__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, TaggedValue):
            return self.value == other.value
        return self.value == other

    def __ne__(self, other):
        return not self.__eq__(other)


class Question(object):
    kind = 'base question'

    def __init__(self,
                 name,
                 message='',
                 choices=None,
                 default=None,
                 ignore=False,
                 validate=True,
                 show_default=False):
        self.name = name
        self._message = message
        self._choices = choices or []
        self._default = default
        self._ignore = ignore
        self._validate = validate
        self.answers = {}
        self.show_default = show_default

    @property
    def ignore(self):
        return bool(self._solve(self._ignore))

    @property
    def message(self):
        return self._solve(self._message)

    @property
    def default(self):
        return self.answers.get(self.name) or self._solve(self._default)

    @property
    def choices_generator(self):
        for choice in self._solve(self._choices):
            yield (
                TaggedValue(*choice)
                if isinstance(choice, tuple) and len(choice) == 2
                else choice
            )

    @property
    def choices(self):
        return list(self.choices_generator)

    def validate(self, current):
        try:
            if self._solve(self._validate, current):
                return
        except errors.ValidationError as e:
            raise e
        except Exception:
            pass
        raise errors.ValidationError(current)

    def _solve(self, prop, *args, **kwargs):
        if callable(prop):
            return prop(self.answers, *args, **kwargs)
        if isinstance(prop, str):
            return prop.format(**self.answers)
        return prop


class Text(Question):
    kind = 'text'

    def __init__(self, name, message='', default=None, **kwargs):
        super(Text, self).__init__(name, message=message,
                                   default=str(default) if default and not callable(default)
                                   else default, **kwargs)


class Password(Text):
    kind = 'password'

    def __init__(self, name, echo='*', **kwargs):
        super(Password, self).__init__(name, **kwargs)
        self.echo = echo


class Editor(Text):
    kind = 'editor'


class Confirm(Question):
    kind = 'confirm'

    def __init__(self, name, default=False, **kwargs):
        super(Confirm, self).__init__(name, default=default, **kwargs)


class List(Question):
    kind = 'list'

    def __init__(self,
                 name,
                 message='',
                 choices=None,
                 default=None,
                 ignore=False,
                 validate=True,
                 carousel=False):

        super(List, self).__init__(
            name, message, choices,
            default, ignore, validate
        )
        self.carousel = carousel


class Checkbox(Question):
    kind = 'checkbox'


# Solution for checking valid path based on
# https://stackoverflow.com/a/34102855/1360532
ERROR_INVALID_NAME = 123


def is_pathname_valid(pathname):
    """
    `True` if the passed pathname is a valid pathname for the current OS;
    `False` otherwise.
    """
    try:
        if not isinstance(pathname, str) or not pathname:
            return False

        _, pathname = os.path.splitdrive(pathname)

        root_dirname = os.environ.get('HOMEDRIVE', 'C:') \
            if sys.platform == 'win32' else os.path.sep

        assert os.path.isdir(root_dirname)
        root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

        for pathname_part in pathname.split(os.path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            except OSError as exc:
                if hasattr(exc, 'winerror'):
                    if exc.winerror == ERROR_INVALID_NAME:
                        return False
                elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                    return False
    except (ValueError, TypeError):
        return False
    else:
        return True


class Path(Text):
    ANY = 'any'
    FILE = 'file'
    DIRECTORY = 'directory'

    kind = 'path'

    def __init__(self, name, default=None, path_type='any', exists=None,
                 normalize_to_absolute_path=False, **kwargs):
        super(Path, self).__init__(name, default=default, **kwargs)

        self._path_type = path_type
        self._exists = exists
        self._normalize_to_absolute_path = normalize_to_absolute_path

        if default is not None:
            try:
                self.validate(default)
            except errors.ValidationError:
                raise ValueError("Default value '{}' is not valid based on "
                                 "your Path's criteria".format(default))

    def validate(self, current):
        super(Path, self).validate(current)

        if current is None:
            raise errors.ValidationError(current)

        current = self.normalize_value(current)

        if not is_pathname_valid(current):
            raise errors.ValidationError(current)

        # os.path.isdir and isfile check also existence of the path,
        # which might not be desirable
        if self._path_type == 'file':
            if self._exists is None and os.path.basename(current) == '':
                raise errors.ValidationError(current)
            elif self._exists and not os.path.isfile(current):
                raise errors.ValidationError(current)
            elif self._exists is not None and not self._exists \
                    and os.path.isfile(current):
                raise errors.ValidationError(current)

        elif self._path_type == 'directory':
            if self._exists is None and os.path.basename(current) != '':
                raise errors.ValidationError(current)
            elif self._exists and not os.path.isdir(current):
                raise errors.ValidationError(current)
            elif self._exists is not None and not self._exists \
                    and os.path.isdir(current):
                raise errors.ValidationError(current)

        elif self._exists and not os.path.exists(current):
            raise errors.ValidationError(current)
        elif self._exists is not None and not self._exists \
                and os.path.exists(current):
            raise errors.ValidationError(current)

    def normalize_value(self, value):
        value = os.path.expanduser(value)

        if self._normalize_to_absolute_path:
            value = os.path.abspath(value)

        return value



# -*- coding: utf-8 -*-
from __future__ import print_function

import sys
from blessed import Terminal

from runflare.inquirer import errors
from runflare.inquirer import events
from runflare.inquirer import themes

from runflare.inquirer.render.console._text import Text
from runflare.inquirer.render.console._editor import Editor
from runflare.inquirer.render.console._password import Password
from runflare.inquirer.render.console._confirm import Confirm
from runflare.inquirer.render.console._list import List
from runflare.inquirer.render.console._checkbox import Checkbox
from runflare.inquirer.render.console._path import Path


class ConsoleRender(object):
    def __init__(self, event_generator=None, theme=None, *args, **kwargs):
        super(ConsoleRender, self).__init__(*args, **kwargs)
        self._event_gen = event_generator or events.KeyEventGenerator()
        self.terminal = Terminal()
        self._previous_error = None
        self._position = 0
        self._theme = theme or themes.Default()

    def render(self, question, answers=None):
        question.answers = answers or {}

        if question.ignore:
            return question.default

        clazz = self.render_factory(question.kind)
        render = clazz(question,
                       terminal=self.terminal,
                       theme=self._theme,
                       show_default=question.show_default)

        self.clear_eos()

        try:
            return self._event_loop(render)
        finally:
            print('')

    def _event_loop(self, render):
        try:
            while True:
                self._relocate()
                self._print_status_bar(render)

                self._print_header(render)
                self._print_options(render)

                self._process_input(render)
                self._force_initial_column()
        except errors.EndOfInput as e:
            self._go_to_end(render)
            return e.selection

    def _print_status_bar(self, render):
        if self._previous_error is None:
            self.clear_bottombar()
            return

        self.render_error(self._previous_error)
        self._previous_error = None

    def _print_options(self, render):
        for message, symbol, color in render.get_options():
            if hasattr(message, 'decode'):  # python 2
                message = message.decode('utf-8')
            self.print_line(' {color}{s} {m}{t.normal}',
                            m=message, color=color, s=symbol)

    def _print_header(self, render):
        base = render.get_header()

        header = (base[:self.width - 9] + '...'
                  if len(base) > self.width - 6
                  else base)
        default_value = ' ({color}{default}{normal})'.format(
            default=render.question.default,
            color=self._theme.Question.default_color,
            normal=self.terminal.normal
        )
        show_default = render.question.default and render.show_default
        header += default_value if show_default else ''
        msg_template = "{t.move_up}{t.clear_eol}{tq.brackets_color}["\
                       "{tq.mark_color}?{tq.brackets_color}]{t.normal} {msg}"

        # ensure any user input with { or } will not cause a formatting error
        escaped_current_value = (
            str(render.get_current_value())
            .replace('{', '{{')
            .replace('}', '}}')
        )
        self.print_str(
            '\n%s: %s' % (msg_template, escaped_current_value),
            msg=header,
            lf=not render.title_inline,
            tq=self._theme.Question)

    def _process_input(self, render):
        try:
            ev = self._event_gen.next()
            if isinstance(ev, events.KeyPressed):
                render.process_input(ev.value)
        except errors.ValidationError as e:
            self._previous_error = e.value
        except errors.EndOfInput as e:
            try:
                render.question.validate(e.selection)
                raise
            except errors.ValidationError as e:
                self._previous_error = render.handle_validation_error(e)

    def _relocate(self):
        print(self._position * self.terminal.move_up, end='')
        self._force_initial_column()
        self._position = 0

    def _go_to_end(self, render):
        positions = len(list(render.get_options())) - self._position
        if positions > 0:
            print(self._position * self.terminal.move_down, end='')
        self._position = 0

    def _force_initial_column(self):
        self.print_str('\r')

    def render_error(self, message):
        if message:
            symbol = '>> '
            size = len(symbol) + 1
            length = len(message)
            message = message.rstrip()
            message = (message
                       if length + size < self.width
                       else message[:self.width - (size + 3)] + '...')

            self.render_in_bottombar(
                '{t.red}{s}{t.normal}{t.bold}{msg}{t.normal} '
                .format(msg=message, s=symbol, t=self.terminal)
            )

    def render_in_bottombar(self, message):
        with self.terminal.location(0, self.height - 2):
            self.clear_eos()
            self.print_str(message)

    def clear_bottombar(self):
        with self.terminal.location(0, self.height - 2):
            self.clear_eos()

    def render_factory(self, question_type):
        matrix = {
            'text': Text,
            'editor': Editor,
            'password': Password,
            'confirm': Confirm,
            'list': List,
            'checkbox': Checkbox,
            'path': Path,
        }

        if question_type not in matrix:
            raise errors.UnknownQuestionTypeError()
        return matrix.get(question_type)

    def print_line(self, base, lf=True, **kwargs):
        self.print_str(base + self.terminal.clear_eol(), lf=lf, **kwargs)

    def print_str(self, base, lf=False, **kwargs):
        if lf:
            self._position += 1

        print(base.format(t=self.terminal, **kwargs), end='\n' if lf else '')
        sys.stdout.flush()

    def clear_eos(self):
        print(self.terminal.clear_eos(), end='')

    @property
    def width(self):
        return self.terminal.width or 80

    @property
    def height(self):
        return self.terminal.width or 24



from distutils.core import setup
import os
import setuptools

from runflare import VERSION

setup(
    name='runflare',
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    version=VERSION,
    description='Official CLI for runflare PaaS',
    author='RUNFLARE DEVELOPMENT TEAM',
    install_requires=[
        'certifi==2021.10.8',
        'charset-normalizer==2.0.7',
        'colorama==0.4.4',
        'fire==0.4.0',
        'halo==0.0.31',
        'idna==3.2',
        'log-symbols==0.0.14',
        'Pygments==2.10.0',
        "blessed==1.19.0",
        'readchar==2.0.1',
        'python-editor==1.0.4',
        'pytz==2021.3',
        'regex==2021.10.8',
        'requests==2.26.0',
        'requests-toolbelt==0.9.1',
        'six==1.16.0',
        'spinners==0.0.24',
        'termcolor==1.1.0',
        'tzdata==2021.2.post0',
        'tzlocal==3.0',
        'urllib3==1.26.7',
        'wcwidth==0.2.5',
        'websockets==10.0',
        'pathspec==0.9.0',
    ],
    author_email='parswebserver@gmail.com',
    url='https://github.com/parswebserver/runflare-cli',
    keywords=['PaaS', 'runflare','Kubernetes'],
    package_data={'.runflare_ignore': ['runflare/.runflare_ignore']},
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Utilities"
    ],
    entry_points='''
        [console_scripts]
        runflare=runflare.runflare:run
    ''',
)
